/*

*/

package language5;

import tree.*;
import tokenizer.*;

/**
   Evaluate (interpret) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.
   <pre>{@code
       Prog ::= Expr
              | '(' 'prog' (Fun | Expr)* Expr ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Expr ')' // formal parameters followed by function body

       Expr ::= Apply
              | If
              | While
              | Begin
              | Var
              | Set
              | Print
              | AExpr
              | BExpr
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' 'apply' Expr Expr* ')'  // function value followed by actual parameters

         If ::= '(' 'if' Expr Expr Expr ')'

      While ::= '(' 'while' Expr Expr ')'

      Begin ::= '(' 'begin' Expr+ ')'

        Var ::= '(' 'var' VARIABLE Expr ')'

        Set ::= '(' 'set' VARIABLE Expr ')'

      Print ::= '(' 'print' Expr ')'

      BExpr ::= '(' '||'  Expr Expr+ ')'
              | '(' '&&'  Expr Expr+ ')'
              | '(' '!'   Expr ')'
              | '('  EqOp Expr Expr ')'
              | '(' RelOp Expr Expr ')'

       EqOp ::= '==' | '!='
      RelOp ::= '<' | '>' | '<=' | '>='

      AExpr ::= '(' '+' Expr Expr* ')'
              | '(' '-' Expr Expr? ')'
              | '(' '*' Expr Expr+ ')'
              | '(' '/' Expr Expr  ')'
              | '(' '%' Expr Expr  ')'
              | '(' '**' Expr Expr ')'

    INTEGER ::= [-|+][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
   }</pre>
*/
public class EvaluateAST_no_error_checking
{
   public static int DEBUG = 1;

   public static Environment globalEnv; // Reference to the global environment object.
                                        // This is needed to implement lexical scope.

   /**
      The methods evaluateProg(), evaluateExpr(), evaluateAexpr(), evaluateBexpr(),
      and evaluateRexpr() are essentially a post-order traversal of the abstract
      syntax tree.

      @param tree  an expression {@link Tree} to be evaluated
      @return the {@link Value} of the expression {@link Tree}
   */
   public static Value eval(final Tree<Token> tree)
   {
      // Instantiate a global environment object.
      globalEnv = new Environment();

      return evaluateProg( tree, globalEnv );
   }


   /**
      Evaluate a program using the given {@link Environment}.

      @param tree  an expression {@link Tree} to be evaluated
      @param env   an {@link Environment} for local variables
      @return the {@link Value} of the expression {@link Tree}
   */
   public static Value evaluateProg(final Tree<Token> tree,
                                    final Environment env)
   {
      final Value result;  // a blank final

      // Check which kind of Prog we have.
      if ( ! tree.root().lexeme.equals("prog") )
      {
         // Evaluate the single expression.
         result = evaluateExpr( tree, env );
      }
      else
      {
         // Evaluate each Fun or Exp in the Prog.
         // Any Fun will have the side effect of putting
         // a function name in the global environment.
         // Any Var expressions will have the side effect
         // of putting a variable in the environment chain.
         // Any Set expressions will have the side effect
         // of changing a value in the environment chain.
         // Any Print expressions will have the side effect
         // of printing an output.
         // Any other expressions would be pointless!
         for (int i = 0; i < tree.subTrees().size()-1; ++i)
         {
            if ( tree.subTrees().get(i).root().lexeme.equals("fun") )
            {
               handleFun( tree.subTrees().get(i), env );
            }
            else
            {
               evaluateExpr( tree.subTrees().get(i), env );
            }
         }

         // Evaluate the last expression and use its
         // value as the value of the prog expression.
         result = evaluateExpr( tree.subTrees().get(tree.subTrees().size()-1), env );
      }

      return result;
   }


   /**
      Handle a function definition. Notice that the return type
      is void since Fun isn't an expression.

      This method mutates the global environment object.
      The Value object put into the environment by this method
      has the tag "lambda" and its value field is a reference
      to the function's "lambda expression".

      @param tree  an expression {@link Tree} for a function definition
      @param env   an {@link Environment} to hold the function definition
   */
   public static void handleFun(Tree<Token> tree, Environment env)
   {
      // Get the function name.
/*1*/ final String name = tree.subTrees().get(0).root().lexeme;

      // Get the "lambda expression" and create a function Value.
/*5*/ final Value lambda = new Value( tree.subTrees().get(1) );

      // Add the <name, value> pair to the global environment.
/*6*/ env.add(name, lambda);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes
   }


   // Evaluate an expression.
   private static Value evaluateExpr(final Tree<Token> tree,
                                     final Environment env)
   {
      final Value result;  // a blank final

      final String node = tree.root().lexeme;

      if ( node.equals("apply") )
      {
         result = evaluateApply( tree, env );
      }
      else if ( node.equals("if") )
      {
         result = evaluateIf( tree, env );
      }
      else if ( node.equals("while") )
      {
         result = evaluateWhile( tree, env );
      }
      else if ( node.equals("begin") )
      {
         result = evaluateBegin(tree, env);
      }
      else if ( node.equals("set") )
      {
         result = evaluateSet(tree, env);
      }
      else if ( node.equals("var") )
      {
         result = evaluateVar(tree, env);
      }
      else if ( node.equals("print") )
      {
         result = evaluatePrint(tree, env);
      }
      else if ( node.equals("&&")
             || node.equals("||")
             || node.equals("!") )
      {
         result = evaluateBexpr(tree, env);  // boolean expression
      }
      else if ( node.equals("==")
             || node.equals("!=") )
      {
         result = evaluateEqexpr(tree, env);  // equality operator
      }
      else if ( node.equals("<")
             || node.equals(">")
             || node.equals("<=")
             || node.equals(">=") )
      {
         result = evaluateRelexpr(tree, env);  // relational operator
      }
      else if ( node.equals("+")
             || node.equals("-")
             || node.equals("*")
             || node.equals("/")
             || node.equals("%")
             || node.equals("**") )
      {
         result = evaluateAexpr(tree, env);  // arithmetic expression
      }
      else if ( tree.subTrees().size() == 0 ) // a leaf node
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( node.equals("true") );
         }
         else if ( node.matches("^[-|+]*[0-9][0-9]*") )
         {
            result = new Value( Integer.parseInt( node ) );
         }
         else if ( env.defined(node) )  // a variable
         {
            result = env.lookUp( node );
         }
         else  // runtime check
         {
            throw new EvalError("undefined variable: "
                                + node
                                + " at line " + tree.root().line
                                + ", position " + tree.root().position);
         }
      }
      else
      {
         throw new EvalError("invalid expression: " + toString(tree));
      }

      return result;
   }


   /**
      This method "applies" a function value to actual parameters.
      This method evaluates the body of the function in an environment
      that binds the actual parameter values (from this function application)
      to the formal parameters (from the function's lambda expression).
   */
   private static Value evaluateApply(Tree<Token> tree,
                                      Environment env)
   {
      // Evaluate the apply's first parameter to a function value.
/*1*/ final Value funValue = evaluateExpr( tree.subTrees().get(0), env );

      // Get a reference to the function's "lambda expression".
/*3*/ final Tree<Token> lambda = funValue.valueL;

      // Create a new environment object that is "nested"
      // in the global environment ("lexical scope") (or, is
      // "nested" in the current environment ("dynamic scope")).
      // This environment is used to bind actual parameter
      // values to formal parameter names.
      // (Note: Change globalEnv to env for dynamic scope.)
/*5*/ final Environment localEnv = new Environment(globalEnv, "Function Activation");

      // Bind, in the new environment object, each actual
      // parameter value to a formal parameter name.
/*6*/ for (int i = 1; i < tree.subTrees().size(); ++i) // iterate through the actual parameters
      {
         // Evaluate, using the current environment chain
         // (NOT the new local environment object)
         // an actual parameter expression's value.
/*7*/    final Value actualParamValue = evaluateExpr( tree.subTrees().get(i), env );

         // Retrieve, from within the lambda expression,
         // a formal parameter name.
/*8*/    final String formalParamName = lambda.subTrees().get(i-1).root().lexeme;

         // Bind, in the new local environment object, the actual
         // parameter value to the formal parameter name.
/*10*/   localEnv.add(formalParamName, actualParamValue);
      }

      if (DEBUG > 0) System.out.println( localEnv + "\n" ); // for debugging purposes

      // Evaluate the body of the lambda expression using the
      // new environment (which contains the binding of the actual
      // parameter values to the function's formal parameter names).
/*11*/final Value result = evaluateExpr(lambda.subTrees().get(tree.subTrees().size()-1), localEnv);

      // When this method returns, the local Environment
      // object that we created to hold the parameter bindings
      // becomes a garbage object (and will be garbage collected).
      // This is like "popping" a stack frame off of the call stack
      // in Java, C , or C++.

      return result;
   }


   // Evaluate an if-expression.
   private static Value evaluateIf(final Tree<Token> tree,
                                   final Environment env)
   {
      final Value result;  // a blank final

      // Evaluate the boolean condition.
      final Value conditionalExpr = evaluateExpr( tree.subTrees().get(0), env );

      // Evaluate the appropriate branch of the if-expression.
      if ( conditionalExpr.valueB )
      {
         result = evaluateExpr( tree.subTrees().get(1), env );
      }
      else
      {
         result = evaluateExpr( tree.subTrees().get(2), env );
      }

      return result;
   }


   // Evaluate a while-loop expression.
   private static Value evaluateWhile(final Tree<Token> tree,
                                      final Environment env)
   {
      // Evaluate the boolean condition.
      Value conditionalExpr = evaluateExpr( tree.subTrees().get(0), env );

      while ( conditionalExpr.valueB )
      {
         // Evaluate the body of the while-loop (for its side effects).
         evaluateExpr( tree.subTrees().get(1), env );
         // Re-evaluate the boolean condition.
         conditionalExpr = evaluateExpr( tree.subTrees().get(0), env );
      }

      // Always return false for a while-loop expression.
      return new Value( false );
   }


   // Evaluate a begin expression.
   private static Value evaluateBegin(final Tree<Token> tree,
                                      final Environment env)
   {
      // Create a new Environment object chained to (or "nested in")
      // the previous ("outer") environment object.
      final Environment newEnv = new Environment(env, "Local (begin)");

      // Evaluate each sub expression in the begin
      // expression (using the new environment chain).
      // The return value of each expression is
      // discarded, so any expression without a
      // side-effect is worthless.
      for (int i = 0; i < tree.subTrees().size()-1; ++i)
      {
         evaluateExpr( tree.subTrees().get(i), newEnv );
      }

      // Evaluate the last expression and use its
      // value as the value of the begin expression.
      final Value result =
         evaluateExpr(tree.subTrees().get(tree.subTrees().size()-1), newEnv);

      // When this method returns, the local Environment
      // object that we created at the beginning of this method
      // becomes a garbage object (and will be garbage collected).
      // This is like "popping" a stack frame off of the call stack
      // in Java, C , or C++.

      return result;
   }


   // Evaluate a set expression.
   private static Value evaluateSet(final Tree<Token> tree,
                                    final Environment env)
   {
      // get the variable
      final String variable = tree.subTrees().get(0).root().lexeme;

      // check if this variable has already been declared
      if ( ! env.defined(variable) )
      {
         throw new EvalError("undefined variable: " + variable);
      }

      // get, and then evaluate, the expression
      final Tree<Token> expr = tree.subTrees().get(1);
      final Value result = evaluateExpr( expr, env );
      // update this variable in the environment
      env.update(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }


   // Evaluate a var expression.
   private static Value evaluateVar(final Tree<Token> tree,
                                    final Environment env)
   {
      // get the variable
      final String variable = tree.subTrees().get(0).root().lexeme;

      // check if this variable has already been declared
      // in the local environment
      if ( env.definedLocal(variable) )
      {
         throw new EvalError("variable already declared: " + variable);
      }

      // get, and then evaluate, the expression
      final Tree<Token> expr = tree.subTrees().get(1);
      final Value result = evaluateExpr( expr, env );

      // declare the new, local, variable
      env.add(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }


   // Evaluate a print expression.
   private static Value evaluatePrint(final Tree<Token> tree,
                                      final Environment env)
   {
      final Value result = evaluateExpr( tree.subTrees().get(0), env );

      if (DEBUG > 0)
      {
         System.out.println( result );
      }
      else
      {
         System.out.println( result.toSimpleString() );
      }

      return result;
   }


   // Evaluate a boolean expression.
   private static Value evaluateBexpr(final Tree<Token> tree,
                                      final Environment env)
   {
      boolean result = false;

      final String node = tree.root().lexeme;

      Value value = evaluateExpr( tree.subTrees().get(0), env );

      result = value.valueB;

      if ( node.equals("&&") )
      {
         for (int i = 1; i < tree.subTrees().size(); ++i)
         {
            if (result)
            {
               value = evaluateExpr( tree.subTrees().get(i), env );

               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         for (int i = 1; i < tree.subTrees().size(); ++i)
         {
            if (! result)
            {
               value = evaluateExpr( tree.subTrees().get(i), env );

               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         result = ! result;
      }

      return new Value( result );
   }


   // Evaluate an equality expression (which is a boolean expression).
   private static Value evaluateEqexpr(final Tree<Token> tree,
                                       final Environment env)
   {
      boolean result = false;

      final String op = tree.root().lexeme;

      final Value valueL = evaluateExpr( tree.subTrees().get(0), env );
      final Value valueR = evaluateExpr( tree.subTrees().get(1), env );

      if ( op.equals("==") )
      {
         if ( ! valueL.tag.equals(valueR.tag) )
         {
            result = false;
         }
         else if ( valueL.tag.equals(Value.INT_TAG) )
         {
            final int resultL = valueL.valueI;
            final int resultR = valueR.valueI;
            result = resultL == resultR;
         }
         else // boolean data type
         {
            final boolean resultL = valueL.valueB;
            final boolean resultR = valueR.valueB;
            result = resultL == resultR;
         }
      }
      else // the '!=' operator
      {
         if ( ! valueL.tag.equals(valueR.tag) )
         {
            result = true;
         }
         else if ( valueL.tag.equals(Value.INT_TAG) )
         {
            final int resultL = valueL.valueI;
            final int resultR = valueR.valueI;
            result = resultL != resultR;
         }
         else // boolean data type
         {
            final boolean resultL = valueL.valueB;
            final boolean resultR = valueR.valueB;
            result = resultL != resultR;
         }
      }

      return new Value( result );
   }


   // Evaluate a relational expression (which is a boolean expression).
   private static Value evaluateRelexpr(final Tree<Token> tree,
                                        final Environment env)
   {
      boolean result = false;

      final String op = tree.root().lexeme;

      Value valueL = evaluateExpr( tree.subTrees().get(0), env );

      Value valueR = evaluateExpr( tree.subTrees().get(1), env );

      final int resultL = valueL.valueI;
      final int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }

      return new Value( result );
   }


   // Evaluate an arithmetic expression.
   private static Value evaluateAexpr(final Tree<Token> tree,
                                      final Environment env)
   {
      int result = 0;

      final String node = tree.root().lexeme;

      final Value valueL = evaluateExpr( tree.subTrees().get(0), env );

      final int resultL = valueL.valueI;
            int resultR = 0;

      if ( tree.subTrees().size() >= 2 )
      {
         final Value valueR = evaluateExpr( tree.subTrees().get(1), env );

         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.subTrees().size() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.subTrees().size(); ++i)
            {
               final Value temp = evaluateExpr( tree.subTrees().get(i), env );

               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( tree.subTrees().size() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         result = resultL * resultR;

         for (int i = 2; i < tree.subTrees().size(); ++i)
         {
            final Value temp = evaluateExpr( tree.subTrees().get(i), env );

            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         result = resultL % resultR;
      }
      else if ( node.equals("**") )
      {
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }


   /**
      Convert a {@code Tree} of {@link Token}s into a {@link String}.
      <p>
      This method does a pre-order traversal of the {@link Tree}.
      <p>
      This method is used by the evaluation errors.

      @param tree  {@link Tree} of {@link Token}s to convert into a {@link String}
      @return a {@link String} representation of this {@code Tree} of {@link Token}s
   */
   public static String toString(final Tree<Token> tree)
   {
      String result = "";
      if ( tree.isLeaf() )
      {
         result += tree.root().lexeme;
      }
      else
      {
         result += "(" + tree.root().lexeme;
         for ( final Tree<Token> subTree : tree.subTrees() )
         {
            result += " " + toString(subTree);  // recursion
         }
         result += ")";
      }
      return result;
   }
}
